/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */


#include <sys/types.h>
#include <errno.h>

#include <sys/mman.h>
#include <stdint.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <poll.h>
#include <sys/time.h>

#include <fcntl.h>
#include <sys/stat.h>

#include <internal/_Message.h>
#include <internal/_Debug.h>
#include <sstream>
#include <internal/_StringArrayHelper.h>
#include <internal/_System.IO.h>
#include <internal/_MemoryMappedFile.h>
#include <internal/_MemoryRegion.h>
#include <internal/_DMA.h>
#include <internal/_Process.h>
#include <common/COIPerf_common.h>
#include <internal/_Perf.h>

#include <unistd.h>

using namespace string_array_helper;
using namespace std;

#if 0
#define DPRINTF(...) printf(__VA_ARGS__)
#else
#define DPRINTF(...)
#endif

// Close the file descriptor across exec calls.
static int setCLOEXEC(int fd)
{
    int flags = fcntl(fd, F_GETFD);
    if (flags == -1) {
        return -1;
    }
    return fcntl(fd, F_SETFD, flags | FD_CLOEXEC);
}

// Call scif open and then modify the resulting
// file descriptor so that it is closed during an exec call.
static scif_epd_t scif_open_cloexec()
{
    scif_epd_t ep = scif_open();
    if (ep == SCIF_OPEN_FAILED) {
        return ep;
    }

    (void) setCLOEXEC(scif_get_fd(ep));

    return ep;
}

// The minimum message size. "headers" will be this big
#define COI_MIN_SEND_RECV_MESSAGE_SIZE 64

// The page size happens to be 4k as well
#define COI_PAGE_SIZE 4096

// The maximum message size used in scif_send/recv will
// be optimized so that you can send pipeline misc data
// in one shot, assuming a sane number of buffers and dependencies.
#define COI_MAX_SEND_RECV_MESSAGE_SIZE (COI_PAGE_SIZE - 1)

// We can keep some registered memory around for even faster transfers,
#define COI_MAX_REGISTERED_MESSAGE_SIZE (COI_PIPELINE_MAX_IN_MISC_DATA_LEN+4096)

#define SIZEOF_UINT64 8
STATIC_ASSERT( SIZEOF_UINT64 == sizeof(uint64_t) );

// TODO: rename. not sure the name makes sense
#define COI_HEADER_DATA_SIZE (COI_MIN_SEND_RECV_MESSAGE_SIZE - SIZEOF_UINT64)

// Allocate enough space to hold a message with a body of size s
char* Message_t::Allocate(size_t s)
{

    Release();
    m_size = s;
    void* ptr = NULL;
    int status = 0;

    // If the message is real small, we need to send at least
    // COI_MIN_SEND_RECV_MESSAGE_SIZE bytes because all of the recv calls
    // will first expect to receive COI_MIN_SEND_RECV_MESSAGE_SIZE bytes.

    // If the message is smaller than COI_MAX_SEND_RECV_MESSAGE_SIZE then
    // we need to allocate those many bytes + the size of a uint64_t.

    // Otherwise we have to allocated enough *aligned* space so that
    // that space can be registered for DMA transfers. We will "waste"
    // an extra page that probably isn't necessary to waste.

    // TODO - 1) Instead of malloc'ing for smallest msg size, you could make it a char[64] member variable.
    //           You would then have to make sure to call free only if you actually called malloc.
    // TODO - 2) Only posix_memalign "m_size" bytes.
    if( m_size <= COI_HEADER_DATA_SIZE )
    {
        ptr = malloc(COI_MIN_SEND_RECV_MESSAGE_SIZE);
    }
    else if( m_size <= (COI_MAX_SEND_RECV_MESSAGE_SIZE - sizeof(uint64_t)) )
    {
        ptr = malloc(m_size + sizeof(uint64_t));
    }
    else
    {
        status = posix_memalign(&ptr,
            COI_PAGE_SIZE, PAGE_CEIL(m_size)+COI_PAGE_SIZE);
    }

    if (ptr == NULL || status != 0)
    {
        throw (std::bad_alloc());
    }

#ifdef DEBUG
    memset(ptr,0,m_size);
#endif
    // The first few bytes of the buffer, when interpreted as a uint64, will give you the size of the
    // rest of the data being transmitted.
    m_buffer = (char*)ptr;
    *(uint64_t*)m_buffer = m_size;

    return buffer();
}

void Message_t::Release()
{
    if (m_buffer)
    {
        free(m_buffer);
    }
    m_buffer  = NULL;
}

char* Message_t::buffer() const
{
    // return the part after where the size was written
    if( m_size <= ( COI_MAX_SEND_RECV_MESSAGE_SIZE - sizeof(uint64_t) ) )
    {
        return m_buffer + sizeof(m_size);
    }
    else
    {
        return m_buffer + COI_PAGE_SIZE;
    }
}

void Message_t::deep_copy(const Message_t& rhs)
{
    Allocate(rhs.m_size); 
    memcpy(m_buffer, rhs.m_buffer, rhs.m_size);
    m_size = rhs.m_size;
}


//=============================================================================
// _SCIFListener Class -- Start
//=============================================================================
//SCIFListener Constructors
_SCIFListener::_SCIFListener():
        m_port_number(0), m_endpoint(SCIF_OPEN_FAILED),m_initialized(false)
{
    PT_ASSERT( pthread_mutex_init(&m_lock, NULL) );
    int status = BindAndListen(0, 1);
    if( status == - 1)
    {
        throw COI_ERROR;
    }
}

_SCIFListener::_SCIFListener(uint16_t portNumber,int backlog):
    m_port_number(0),m_endpoint(SCIF_OPEN_FAILED),m_initialized(false)
{
    PT_ASSERT( pthread_mutex_init(&m_lock, NULL) );
    int status = BindAndListen(portNumber,backlog);
    if( status == - 1)
    {
        throw COI_ERROR;
    }
}

// Opens a scif endpoint, binds to a port, calls scif_listen.
int
_SCIFListener::BindAndListen(uint16_t portNumber,int backlog)
{
    int status = -1;
    _PthreadAutoLock_t lock( m_lock );
    if(!m_initialized)
    {
        m_endpoint = scif_open_cloexec();

        DPRINTF("_SCIFListener::BindAndListen scif_open %d\n", m_endpoint);
        if( m_endpoint == SCIF_OPEN_FAILED )
        {
            m_initialized = false;
            perror("_SCIFListener::BindAndList: scif_open");
            return -1;
        }

        status = scif_bind(m_endpoint, portNumber);
        if( status == -1 )
        {
            m_initialized=false;
            perror("_SCIFListener::BindAndList: scif_bind");
            return -1;
        }
        m_port_number = (uint16_t)status;

        status = scif_listen(m_endpoint,backlog);
        if( status == -1 )
        {
            m_initialized=false;
            perror("_SCIFListener::BindAndList: scif_listen");
            return -1;
        }
        m_initialized=true;
    }
    return m_port_number;
}

int
_SCIFListener::GetPortNumber()
{
    return m_port_number;
}

// Helper function returns milliseconds elapsed since t0
static long millis_elapsed(struct timeval &t0) {
    struct timeval t1;
    gettimeofday(&t1, NULL);
    return (t1.tv_sec - t0.tv_sec) * 1000 + (t1.tv_usec - t0.tv_usec) / 1000;
}

// This wrapper calls poll, it deals with interruption as well as permitting
// a timeout. It leaves errno set as defined by poll(2). As with poll, it
// never leaves errno ETIMEDOUT, that case is detected by a return value of 0.
//
// poll can be interrupted by various signals including LWP library signals
// and stuff we have no control over. We need to restart the system call in
// such cases. In fact, signal(7) specifies that poll (and many other functions)
// do not restart even when you have a sigaction(2) handler with the SA_RESTART
// flag set.
static int poll_retry(struct scif_pollepd *parr, size_t parr_len, int timeout_ms)
{
    struct timeval t0;
    if( timeout_ms > 0 ){
        // Only make this call if there is a timeout
        gettimeofday(&t0, NULL);
    }
    int timeleft = timeout_ms;

    int n = 0;
    while (1) {
        n = scif_poll(parr, (unsigned int) parr_len, timeleft);
        if (n == -1 && errno == EINTR) {
            // interrupted
            errno = 0;
            if (timeout_ms > 0) {
                // timed poll, are we out of time?
                timeleft -= millis_elapsed(t0);
                if (timeleft < 0) {
                    return 0;
                }
            } // else: untimed poll no work needed, retry.
        } else {
            // * n > 0 (poll got something)
            // * non-EINTR error (n == -1 and errno is something else)
            // * n == 0 (timed out without interruption)
            break;
        }
    }
    return n;
}

int
_SCIFListener::WaitForConnect(_SCIFComm& comm, int timeout_ms)
{

    //TODO something about flags
    scif_epd_t new_Epdt;
    scif_portID_t peerPort;
    int status = -1;

    _PthreadAutoLock_t lock( m_lock );
    if(m_initialized)
    {
        struct scif_pollepd poll_struct;
        poll_struct.epd = m_endpoint;

        // I can't find in the scif documentation what event is signaled
        // when someone calls connect(), so I am going to poll for all of them.
        poll_struct.events = SCIF_POLLIN;
        poll_struct.revents = 0;

        int n = poll_retry(&poll_struct, 1, timeout_ms);
        if (n == 1)
        {
            status = scif_accept(m_endpoint, &peerPort, &new_Epdt, SCIF_ACCEPT_SYNC);
            DPRINTF("scif_accept called on local end point %d, status %d\n", m_endpoint, status);
            if (status == 0) {
                (void) setCLOEXEC(new_Epdt);
            }
        }
        else if (n == 0)
        {
            status = -ETIMEDOUT;
        }
        else
        {
            status = -errno;
        }

        if( status != 0 )
        {
            return status;
        }

        // Initialize will scif_close() new_Epdt if it fails. Otherwise
        // it would need to get done before returning out of this function, or
        // as part of the comm object's destructor.
        status = comm.Initialize(new_Epdt,peerPort);

    }

    return status;
}

int
_SCIFListener::WaitForConnect(_SCIFComm& comm)
{
    // If they haven't connected back to us in 10 seconds, return.
    return WaitForConnect(comm, 10*1000);
}

_SCIFListener::~_SCIFListener()
{
    // No locking here, Close() will take care of it, but mostly
    // because we don't do anything with "status" other than a
    // DPRINTF. If more complex things are done here, the lock
    // will need to be changed to recursive before adding
    // a lock at the beginning of this destructor.
    if( m_initialized )
    {
        int status = Close();
        if( status == -1 )
        {
            DPRINTF("Error closing _SCIFListener\n");
        }
    }
}

// Closes the listener
int
_SCIFListener::Close()
{
    int status = -1;
    _PthreadAutoLock_t lock( m_lock );
    if(m_initialized)
    {
        status = scif_close(m_endpoint);
        if( status == -1 )
        {
            perror("scif_close");
            return -1;
        }
        m_endpoint = SCIF_OPEN_FAILED;
        m_initialized = false;
        m_port_number = 0;
    }
    return status;
}



//=============================================================================
// _SCIFListener Class -- End
//=============================================================================




//=============================================================================
// _SCIFComm Class -- Start
//=============================================================================

#define EXIT "Exit"
#define EXIT_LEN 4

// Default constructor
_SCIFComm::_SCIFComm() : m_port_num(0),
              m_endpoint(SCIF_OPEN_FAILED),
              m_pipe_Recv_endpoint(SCIF_OPEN_FAILED), m_pipe_Send_endpoint(SCIF_OPEN_FAILED),
              m_initialized(false),
              m_send_signal_offset(SCIF_REGISTER_FAILED),
              m_registered_space_offset(SCIF_REGISTER_FAILED),
              m_remote_offset(SCIF_REGISTER_FAILED)
{
    pthread_mutexattr_t   mta;
    pthread_mutexattr_init(&mta);
    pthread_mutexattr_settype(&mta, PTHREAD_MUTEX_RECURSIVE);

    int status = pthread_mutex_init(&m_lock,&mta);

    pthread_mutexattr_destroy(&mta);

    if( status )
    {
        throw COI_ERROR;
    }
}

_SCIFComm::~_SCIFComm()
{
    Disconnect();
}

// An out-of-band (with respect to SCIF) way to close a connection.
// SCIF (and I think standard sockets) don't always notify the remote
// endpoint that they are going away. This also serves as a way to
// forcibly close a SCIF connection.
void _SCIFComm::SendCloseSignal()
{
    if (m_initialized)
    {
        char Exit[] = "Exit";
        if (scif_send(m_pipe_Send_endpoint,Exit,EXIT_LEN,SEND_BLOCK) != EXIT_LEN)
        {
            perror("error occured writing exit signal to _SCIFComm pipe");
        }
    }
}

void *_SCIFComm::_temp_thread(scif_epd_t inEpdtmp)
{
    if ((scif_listen(inEpdtmp, 1)) < 0) {
        goto _ret;
    }

    struct scif_portID tmpPort;
    if ((scif_accept(inEpdtmp, &tmpPort, &m_pipe_Recv_endpoint, SCIF_ACCEPT_SYNC)) < 0)
    {
        goto _ret;
    }

_ret:
    return (void*)0;
}

int _SCIFComm::setupConnection(scif_epd_t inEpd, int portNum)
{
    int rv = 0;
    pthread_t iScifThread;
    uint16_t myNode = 0;

    {
      uint16_t allNodes[32];
      if ((rv=scif_get_nodeIDs(allNodes, sizeof(allNodes)/sizeof(allNodes[0]), &myNode)) <= 0)
    {
        return rv;
    }
    }

    struct _temp_thread_parameter threadParameter;

    threadParameter.mine = this;
    threadParameter.epd  = inEpd;
    /* Create a Thread to Listen and Accept the Connecting Requests */
    if (COI_SUCCESS != pthread_create(&iScifThread, NULL, _temp_thread_static, (void *)&threadParameter))
    {
        return 1;
    }

    if ((m_pipe_Send_endpoint = scif_open()) < 0) {
        rv = 2;
        goto _ret;
     }
     {
         struct scif_portID port;
         int preConnerrno = 0;
         int preConerrnoValid = 0;

         port.node = myNode;
         port.port = portNum;

_retry:
         if ((scif_connect(m_pipe_Send_endpoint, &port)) < 0)
         {
             if (errno == ECONNREFUSED)
             {
                 if (preConerrnoValid && (preConnerrno==ECONNREFUSED))
                     goto _retry;
                 preConnerrno = errno;
                 preConerrnoValid = 1;
                 goto _retry;
             }
             rv = 3;
             goto _ret;
         }
    }
_ret:    
    // Make sure all connection from peers are done
    pthread_join(iScifThread,NULL);
    return rv;
}

int
_SCIFComm::Initialize_pipe_endpoints()
{
    int errInfo = 0;
    if ((m_pipe_Recv_endpoint == SCIF_OPEN_FAILED) &&
        (m_pipe_Send_endpoint == SCIF_OPEN_FAILED))
    {
        scif_epd_t tempepd;
        int con_pn;
        if ((tempepd=scif_open()) < 0) {
            // No joy.  Cannot open the initial endpoint.
            return -1;
        }
        if ((con_pn = scif_bind(tempepd, 0)) < 0)
        {
            return -2;
        }
        setupConnection(tempepd,con_pn);
        scif_close(tempepd);
    }
    return errInfo;
}

int
_SCIFComm::Initialize(scif_epd_t epdt, scif_portID_t port)
{
    int status = -1;
    _PthreadAutoLock_t lock( m_lock );
    if (!m_initialized)
    {
        if (Initialize_pipe_endpoints())
        {
            // Ownership of the endpoint was given to the SCIFComm object so
            // the object will cleanup by closing the endpoint here. Normally
            // this is handled by the destructor/disconnect(), but those
            // check on "m_initialized" before deciding to take any action.
            status = scif_close( epdt );
            assert( status == 0 );
            
            std::stringstream err;
            err << "_SCIFComm::Initialize: Initialize_pipe_endpoints failed on thread " << (void*)pthread_self() << endl;
            perror(err.str().c_str());
            return -1;
        }
        m_endpoint = epdt;
        m_peer_port = port;
        status = 0;
        m_initialized = true;
    }

    return status;
}

int
_SCIFComm::Bind(uint16_t portNumber)
{
    _PthreadAutoLock_t lock( m_lock );
    if (!m_initialized)
    {
        m_endpoint = scif_open_cloexec();
        DPRINTF("_SCIFComm::Bind scif_open returned %d\n", m_endpoint);

        if( m_endpoint == SCIF_OPEN_FAILED )
        {
            perror("_SCIFComm::Bind: scif_open");
            return -1;
        }

        int status = scif_bind(m_endpoint, portNumber);
        if( status == -1 )
        {
             perror("_SCIFComm::Bind: scif_bind");
             return -1;
        }
        m_port_num = (uint16_t)status;

        if (Initialize_pipe_endpoints()) {
            perror ("_SCIFComm::Bind Initialize_pipe_endpoints failed");
            scif_close(m_endpoint);
            m_endpoint = SCIF_OPEN_FAILED;
            return -1;
        }
        m_initialized = true;
    }

    return m_port_num;
}

int
_SCIFComm::Connect(uint16_t node,uint16_t portNumber, bool reconnect)
{
    int status = -1;
    _PthreadAutoLock_t lock( m_lock );
    if( reconnect )
    {
        Disconnect();
        status = Bind(0);
    }

    if (m_initialized)
    {
        m_peer_port.node=node;
        m_peer_port.port=portNumber;

        // We do scif_connect() in a loop. If we encounter ECONNREFUSED,
        // that implies the peer endpoint's backlog is full (or the
        // endpoint is closed), in either case, we do an exponential
        // backoff until we have waited "long enough" to consider the
        // connection a failure.
        useconds_t backoff = 10*1000;
        while (1) {
            status = scif_connect(m_endpoint, &m_peer_port);
            if (status != -1 || errno != ECONNREFUSED) {
                break;
            }

            if (usleep(backoff) == -1 && errno == EINTR) 
            {
                break; // bail if we catch a signal
            }
            else if (backoff < 32*1000*1000)
            {
                backoff *= 2;
            }
            else
            {
                break; // it's been too long
            }

        }
    }

    // Older scif drivers return 0 from scif_connect, so a successful status
    // may or may not be 0.
    // It is possible, but not required that Bind() was called
    // before calling Connect(), so m_port_num may or may not be 0.

    // A branch for the old version of the scif driver
    if( status == 0 )
    {
        //No way of knowing what port we are bound to?
        return 0;
    }

    if( status > 0 )
    {
#if DEBUG
        // sanity check for "if we...err, the comm object, was already bound to some port
        // did scif_connect randomly rebind it to a diff port"
        if(m_port_num > 0)
        {
            assert( m_port_num == (uint16_t)status );
        }
#endif
        m_port_num = (uint16_t)status;

        return 0;
    }

    return -1;
}


void
_SCIFComm::ForkDisconnect()
{
     //This function explicity skips the scif_unregister, since
     //its entire purpose is to prevent SCIF connections living
     //through a fork. SCIF should auto-cleanup its own
     //registered space if this is the last reference.
     //Infact we can't cleanup registered
     //space as that might unregister the address range from the
     //parent process, which is highly dangerous.
     if (m_initialized)
     {
        m_initialized = false;

        //Later we might want to add cleanup for the memory maps
        //but nothing from the scif endpoints.

        if( m_send_signal_offset != SCIF_REGISTER_FAILED )
        {
            m_send_signal_offset = SCIF_REGISTER_FAILED;
            munmap((void*)m_send_signal_page, PAGE_SIZE);
        }

        if( m_registered_space_offset != SCIF_REGISTER_FAILED )
        {
            m_registered_space_offset = SCIF_REGISTER_FAILED;
            munmap(m_registered_space, COI_MAX_REGISTERED_MESSAGE_SIZE);
        }

        m_remote_offset = SCIF_REGISTER_FAILED;

        scif_close(m_pipe_Recv_endpoint);
        m_pipe_Recv_endpoint = SCIF_OPEN_FAILED;

        scif_close(m_pipe_Send_endpoint);
        m_pipe_Send_endpoint = SCIF_OPEN_FAILED;

        scif_close(m_endpoint);
        m_endpoint = SCIF_OPEN_FAILED;
     }
}



// Disconnection the connection, set stuff to well-known "-1" values.
int
_SCIFComm::Disconnect()
{
     int status=-1;
     _PthreadAutoLock_t lock( m_lock );
     if (m_initialized)
     {
        m_initialized = false;

        if( m_send_signal_offset != SCIF_REGISTER_FAILED )
        {
            scif_unregister(m_endpoint, m_send_signal_offset, PAGE_SIZE);
            m_send_signal_offset = SCIF_REGISTER_FAILED;
            munmap((void*)m_send_signal_page, PAGE_SIZE);
        }

        if( m_registered_space_offset != SCIF_REGISTER_FAILED )
        {
            scif_unregister(m_endpoint, m_registered_space_offset, COI_MAX_REGISTERED_MESSAGE_SIZE);
            m_registered_space_offset = SCIF_REGISTER_FAILED;
            munmap(m_registered_space, COI_MAX_REGISTERED_MESSAGE_SIZE);
        }

        m_remote_offset = SCIF_REGISTER_FAILED;

        // TODO: check the return from close() and log it
        scif_close(m_pipe_Recv_endpoint);
        m_pipe_Recv_endpoint = SCIF_OPEN_FAILED;
        
        scif_close(m_pipe_Send_endpoint);
        m_pipe_Send_endpoint = SCIF_OPEN_FAILED;

        status = scif_close(m_endpoint);
        m_endpoint = SCIF_OPEN_FAILED;
    }

    return status;
}

// Send an array of arrays by first converting it to a single array that
// is NULL terminated. count indicates that maximum number of entries
// in the array you wish to send. "array" isn't trusted.
COIRESULT _SCIFComm::SendStringArrayUnsafe(const char** array, uint32_t count)
{
    Message_t message;
    uint64_t size = 0;

    COIRESULT result;

    string_vector::get_max_count_and_size(array, count, UINT_MAX, count, size);
    DPRINTF("String array has %u elements with total size = %lu\n", count, size);
    // We don't support 4 GiB of string arrays. We just don't.
    if( size > UINT_MAX )
    {
        return COI_OUT_OF_RANGE;
    }

    message.Allocate(size);
    char* buffer = message.buffer();
    memset( buffer, 0, size );

    for( uint32_t i = 0; i < count; i++ )
    {
        // strcpy and strlen are safe because of the get_max_count_and_size checks above
        strcpy(buffer, array[i]);
        DPRINTF("string = %s\n", buffer);
        buffer += (strlen(buffer) + 1);
    }
    DPRINTF("\n");

    result = SendUnsafe( message );
    return result;
}

COIRESULT _SCIFComm::SendFilesAndRecvResponseUnsafe(vector<string>& file_names)
{
    COIRESULT result;

    vector<MemoryMappedFile*> files;
    vector<void*> buffers;
    vector<uint64_t> lengths;

    // Send the files corresponding to the names we just sent
    for(vector<string>::iterator i = file_names.begin(); i != file_names.end(); i++ )
    {
        MemoryMappedFile* file =  new MemoryMappedFile(i->c_str(), "rb", PROT_READ, MAP_PRIVATE, 0);

        if( !file->IsGood() )
        {
            break;
        }

        files.push_back(file);
        buffers.push_back(file->Buffer);
        lengths.push_back(file->GetLength());
    }

    // Let the function called below catch any errors
    result = SendFileBuffersAndRecvResponseUnsafe(file_names, buffers, lengths);

    while(!files.empty())
    {
        MemoryMappedFile* file = files.back();
        delete file;
        file = NULL;
        files.pop_back();
    }

    return result;
}

COIRESULT _SCIFComm::SendFileBuffersAndRecvResponseUnsafe(vector<string>&      names,
                                                    vector<void*>&       buffers,
                                                    vector<uint64_t>&    lengths)
{
    COIRESULT result;
    if(!m_initialized)
    {
        return COI_NOT_INITIALIZED;
    }

    if( names.size() != buffers.size() || names.size() != lengths.size() )
    {
        return COI_ARGUMENT_MISMATCH;
    }

    // Convert vector<string> into string_vector
    string_vector files;
    files.insert(files.begin(), names.begin(), names.end() );
    // Send the list of the file names.
    result = SendStringArrayUnsafe( files );
    if( result != COI_SUCCESS )
    {
        return result;
    }

    // Send the files corresponding to the names we just sent
    vector<void*>::iterator buffers_iter;
    vector<uint64_t>::iterator lengths_iter;
    for(    buffers_iter = buffers.begin(), lengths_iter = lengths.begin();
            buffers_iter != buffers.end();
            buffers_iter++, lengths_iter++)
    {
        Message_t file_msg;

        file_msg.Allocate( *lengths_iter );
        memcpy( file_msg.buffer(), *buffers_iter, *lengths_iter );

        result =  SendUnsafe(file_msg);
        if( result != COI_SUCCESS )
        {
            return result;
        }
        DPRINTF("Sent %s to the device\n", names[buffers_iter - buffers.begin()].c_str());
    }

    // Recv the response
    Message_t response;
    result = ReceiveUnsafe( response );
    if( result != COI_SUCCESS )
    {
        return result;
    }

    result = *((COIRESULT*)response.buffer());

    return result;

}

COIRESULT  _SCIFComm::ReceiveFiles(const std::string& base_dir,
    /*out*/ vector<string>& files_written_with_path,
    /*out*/ vector<string>& files_written_original)
{
    if(!m_initialized) {
        DPRINTF("ReceiveFiles on uninitialized _SCIFComm\n");
        return COI_ERROR;
    }

    COIRESULT result = COI_ERROR;

    // Recv the list of files about to be transferred
    DPRINTF("Receive the list of files about to be transferred\n");
    Message_t files_msg;
    vector<string> files_to_write;
    result = ReceiveUnsafe( files_msg );
    if( result != COI_SUCCESS )
    {
        return result;
    }
    DPRINTF("Received %lu bytes\n", files_msg.size());

    string_vector::add(files_to_write, static_cast<char*>(files_msg.buffer()),
        (uint32_t)files_msg.size() );
    DPRINTF("Files to write size are %lu\n", files_to_write.size());
    size_t written = 0;

    // Now receive the files.
    // Each file will be written into <base_dir>/FileNameOnlyOf( file to write ).
    // Before writing the file it will be removed if it exists.
    // This is preferable to a regular "overwrite" because in memory filesystems
    // you can overwrite the contents of a file that is currently in use and
    // that will lead to badness.
    for (vector<string>::iterator i = files_to_write.begin();
        i != files_to_write.end(); i++ )
    {
        DPRINTF("Receiving file %s\n", i->c_str());
        Message_t file_msg;
        result = ReceiveUnsafe( file_msg );
        if( result != COI_SUCCESS ) return result;

        string filename_only;
        System::IO::Path::GetFile(*i, filename_only);
        string filename;
        int status = System::IO::Path::Combine(base_dir, filename_only,
            filename);
        if (status != -1)
        {
            DPRINTF("Writing file %s...", filename.c_str());
            bool success = System::IO::File::UnlinkAndWrite(filename,
                                                            file_msg.buffer(),
                                                            file_msg.size());
            if (success)
            {
                written++;
                files_written_original.push_back(*i);
                files_written_with_path.push_back(filename);
                DPRINTF("Written.");
            }
            DPRINTF("\n");
        }
    }
    DPRINTF("Wrote %lu of %lu files\n", written, files_to_write.size());
    result = files_to_write.size() == written ? COI_SUCCESS : COI_ERROR;

    return result;
}

COIRESULT _SCIFComm::ScifSendWrapper(void* buffer, size_t size)
{
    if( !buffer )
    {
        return COI_INVALID_POINTER;
    }
    if( size > INT_MAX )
    {
        return COI_OUT_OF_RANGE;
    }

    const int size_as_int = (int)size;
    int bytes_sent = 0;

    while ( bytes_sent < size_as_int )
    {
        // TODO: Now that SCIF_SEND_BLOCK doesn't guarantee all the bytes were sent
        //       we may as well replace it with a non-blocking send. But that can consume
        //       CPU cycles, which we must take care of not spinning forever if that's the case.
        //       The usual "spin for a while before you move to an async version" should be done here.
        int status = scif_send( m_endpoint, buffer, (size_as_int - bytes_sent), SCIF_SEND_BLOCK );

        if( status == -1 )
        {
            // there was some failure
            if(IsErrnoCOI_PROCESS_DIED(errno) )
            {
                return COI_PROCESS_DIED;
            }
            else
            {
                return COI_ERROR;
            }
        }
        else if( status > 0 )
        {
            // we sent some bytes. even though we requested blocking send,
            // scif was changed such that scif_send may not send all the bytes.
            bytes_sent += status;

            // Sanity check that scif_send didn't do something real weird
            assert( bytes_sent <= size_as_int );
        }
        else
        {
            // "should never happen"
            assert( 0 );
            return COI_ERROR;
        }
    }

    return COI_SUCCESS;

}
// The algorithm for receiving size bytes is to try to receive them
// by burning CPU cycles in an async recv loop. However, we cannot have
// threads burning CPU cycles forever, so we have a limit on how long we
// do that before switching to a sync/blocking way to recv data that doesn't
// burn CPU cycles.
COIRESULT _SCIFComm::ScifRecvWrapper(void* buffer, size_t size)
 {

    if(!buffer)
    {
        return COI_INVALID_POINTER;
    }

    struct scif_pollepd poll_structs[2];

    poll_structs[0].epd = m_pipe_Recv_endpoint;
    poll_structs[0].events = SCIF_POLLIN;
    poll_structs[0].revents = 0;

    poll_structs[1].epd = m_endpoint;
    poll_structs[1].events = SCIF_POLLIN;
    poll_structs[1].revents = 0;

    if(poll_structs[1].epd < 0)
    {
        return COI_ERROR;
    }

    //Lame... This has to be here for now, as for some
    //reason this is the FIRST time it is called druing process creation
    //and if it is NOT present the whole program segfaults.
    //This needs to be corrected, and moved someplace else
    //thinking COIEngineGetInfo, really hoping this is not a product
    //of the test content linking in both the static common and dynamic libs
    SYMBOL_VERSION(COIPerfGetCycleFrequency,1)();

    int status = 0;
    size_t bytes_read = 0;
    char* bufferPtr = (char*)buffer;
    DPRINTF("Start to recv %lu bytes total\n", size);

    while (bytes_read < size)
    {
        static const uint64_t max_wait = COIDMAFence::MAXSPINTIME * 1000000;
        struct timeval t0, t1;
        gettimeofday(&t0, NULL);
        uint64_t elapsed = 0;

        do
        {
            status = scif_recv( m_endpoint, bufferPtr, (int) (size - bytes_read), 0);
            gettimeofday(&t1, NULL);
            elapsed = (t1.tv_sec - t0.tv_sec) * 1000000 + (t1.tv_usec - t0.tv_usec);
        } while( status == 0 && (elapsed < max_wait) );

        // if any bytes were read
        if (status > 0)
        {
            bytes_read += status;
            bufferPtr += status;
            DPRINTF("Received %d bytes. Running Total is %lu"
                    "(%lu total expected)\n", status, bytes_read, size);
        }
        // if no bytes were available to read we call "poll" in a blocking fashion
        // so that we don't spend CPU cycles in a busy loop any more.
        // We could have just used the blocking version of scif_recv, but we found
        // out that it had certain limitations and we were better served by calling
        // poll() and then if there was data to call scif_recv().
        else if( status == 0 )
        {
            DPRINTF("No bytes available. Going to poll.  errno= %d\n", errno);
            poll_retry(poll_structs, 2, -1);
            DPRINTF("Poll completed. There is data available to scif_recv or there is an error. errno= %d\n", errno);
#ifdef DEBUG
            for(int i=0; i<2; i++)
            {
                if( poll_structs[i].revents )
                {
                    DPRINTF("poll structs[ %d ] was signaled. value = 0x%x\n",
                            i, poll_structs[i].revents);
                }
            }
#endif
            // If the 0th descriptor was set for read the socket is
            // trying to close connection. Exit with a error so that socket
            // can disconnect
            if (((poll_structs[0].revents & SCIF_POLLIN) == SCIF_POLLIN))
            {
                char buf[EXIT_LEN];
                scif_recv(poll_structs[0].epd, buf, EXIT_LEN, 0);
                return COI_ERROR;
            }
        }
        // scif_recv returned an error
        else
        {
            DPRINTF("scif_recv got an error. errno = %s\n", strerror(errno));


            if ( IsErrnoCOI_PROCESS_DIED(errno) )
            {
                return COI_PROCESS_DIED;
            }

            return COI_ERROR;
        }
    } //end while
    if (bytes_read == size)
    {
        return COI_SUCCESS;
    }
    else
    {
        return COI_ERROR;
    }
 }

// returns success if a signal page is
// successfully allocated and registered
COIRESULT _SCIFComm::AllocateSignalPage()
{
    if ( m_send_signal_offset != SCIF_REGISTER_FAILED )
    {
        return COI_ALREADY_INITIALIZED;
    }
    m_send_signal_page = (uint64_t*)mmap(0, PAGE_SIZE, PROT_READ|PROT_WRITE,
                                         MAP_ANON|MAP_SHARED, -1, 0);
    if( m_send_signal_page == MAP_FAILED)
    {
        return COI_OUT_OF_MEMORY;
    }
    m_send_signal_offset = scif_register(m_endpoint, (void*)m_send_signal_page,
                                         PAGE_SIZE, 0,
                                         SCIF_PROT_READ|SCIF_PROT_WRITE, 0);
    if( m_send_signal_offset == SCIF_REGISTER_FAILED )
    {
        DPRINTF("signal registration failed %d\n", errno);
        UNUSED_ATTR int status = munmap((void*)m_send_signal_page, PAGE_SIZE);
        assert(status);
        m_send_signal_page = (uint64_t*)MAP_FAILED;
        return COI_ERROR;
    }

    DPRINTF("Allocated and registered the signal page 0x%lx\n", m_send_signal_offset);
    return COI_SUCCESS;
}

// Ensures that m_registered_space and m_registered_space_offset have been
// allocated and scif_registered, or returns an error otherwise.
COIRESULT _SCIFComm::AllocateRegisteredMemory()
{
    if( m_registered_space_offset != SCIF_REGISTER_FAILED )
    {
        return COI_ALREADY_INITIALIZED;
    }
    m_registered_space = mmap(0, COI_MAX_REGISTERED_MESSAGE_SIZE,
                              PROT_READ|PROT_WRITE, MAP_ANON|MAP_SHARED,
                              -1, 0);
    if( m_registered_space == MAP_FAILED)
    {
        return COI_OUT_OF_MEMORY;
    }
    m_registered_space_offset = scif_register(m_endpoint, m_registered_space,
                                              COI_MAX_REGISTERED_MESSAGE_SIZE,
                                              0, SCIF_PROT_READ|SCIF_PROT_WRITE, 0);
    if( m_registered_space_offset == SCIF_REGISTER_FAILED )
    {
        UNUSED_ATTR int status = munmap(m_registered_space, COI_MAX_REGISTERED_MESSAGE_SIZE);
        assert(status);
        m_registered_space = MAP_FAILED;
        return COI_ERROR;
    }
    DPRINTF("Allocated and registered memory for DMA %p\n", (void*) m_registered_space_offset);
    return COI_SUCCESS;

}

// TODO: Ended up not needing the "offset" field.
//       Remove it.
// A way to represent the smallest Message_t in C and use these
// as stack vars instead of calling Allocate().
typedef struct Header_t
{
    uint64_t size;
    union
    {
        int   scif_write_status;
        off_t offset;
        char data[COI_HEADER_DATA_SIZE];
    };
} Header_t;

STATIC_ASSERT( sizeof(Header_t) == COI_MIN_SEND_RECV_MESSAGE_SIZE );

COIRESULT _SCIFComm::SendUnsafe(Message_t& message_to_send)
{
    if (!m_initialized)
    {
        return COI_NOT_INITIALIZED;
    }
    if (m_endpoint == SCIF_OPEN_FAILED)
    {
        return COI_OUT_OF_RANGE;
    }

    int status = -1;
    COIRESULT result;
    const uint64_t message_size = message_to_send.size();
    const uint64_t raw_size = message_size + sizeof(uint64_t);

    if (raw_size <= COI_MIN_SEND_RECV_MESSAGE_SIZE)
    {
        // if the message is really small we would be sending
        // a struct kinda like
        // {
        //     uint64_t size;
        //     char    rest_of_data[];
        // }
        // All of the COI_MIN_SEND_RECV_MESSAGE_SIZE bytes are sent using scif_send
        DPRINTF("Sending %d: sizeof(int64_t) + %lu bytes + "
            "padded into 64 bytes.\n", m_endpoint, message_size);
        result = ScifSendWrapper(message_to_send.rawdata(),
                                 COI_MIN_SEND_RECV_MESSAGE_SIZE);
        return result;
    }
    else if (raw_size <= COI_MAX_SEND_RECV_MESSAGE_SIZE)
    {
        // if the message is small it's mostly the same as really small,
        // but the size is a little diff
        // a struct kinda like
        // {
        //     uint64_t size;
        //     char     rest_of_data[];
        // }
        // All of the raw_size bytes are sent using scif_send
        DPRINTF("Sending sizeof(int64_t) + %lu bytes of the actual message\n",
            message_size);
        result = ScifSendWrapper(message_to_send.rawdata(),
                                 raw_size);
        return result;
    }
    else if( message_size <= COI_MAX_REGISTERED_MESSAGE_SIZE )
    {
        // if the message is not small, the current layout of the data looks like
        // {
        //      uint64_t size; //aligned to PAGE_SIZE
        //      int      scif_write_status;
        //      char     unused_data[COI_MIN_SEND_RECV_MESSAGE_SIZE - sizeof( two things above) ];
        // }
        // Always:
        //   Send a small message with the size being transferred padded into COI_MIN_SEND_RECV_MESSAGE_SIZE bytes
        // If first time a message of this size is sent, then:
        //   allocate and scif_register memory
        //   recv the remote side's scif_offset for the destination of the DMA
        //   send the scif_offset for our memory registered for DMA
        // Always:
        //   memcpy message contents into local memory registered for DMA
        //   DMA the contents to the device using scif_writeto
        //   wait for the DMA to complete
        //   send a message to the receiver telling them the host has finished the DMA
        //   wait for the receiver to send a message telling us that the have finished copying data out of the buffer
        //

        // For faster transfers you will need a signal page to use with scif fence and busy polling.
        // Allocate the signal page if necessary and register it.
        // Either both succeed or both must be initialized to their "fail" states.
        if( message_size <= COIDMAFence::MAXSPINSIZE && m_send_signal_offset == SCIF_REGISTER_FAILED )
        {
            result = AllocateSignalPage();
            if( result != COI_SUCCESS )
            {
                return result;
            }
        }

        Header_t* header = (Header_t*)message_to_send.rawdata();

        // If this is the first time a message in this size range is being sent we will
        // need to allocate and scif_register some memory.
        // Allocate the registered memory where we'll copy stuff to temporarily, if needed
        if( m_registered_space_offset == SCIF_REGISTER_FAILED )
        {

            // Let the other side know about the message
            result = ScifSendWrapper(header, sizeof(Header_t));
            if( result != COI_SUCCESS )
            {
                return COI_ERROR;
            }

            result = AllocateRegisteredMemory();
            if( result != COI_SUCCESS )
            {
                return result;
            }
            // If we just allocated stuff that they just did too
            // and we'll also need the remote offset.
            result = ScifRecvWrapper(&m_remote_offset, sizeof(m_remote_offset));
            if( result != COI_SUCCESS )
            {
                return result;
            }
            if( m_remote_offset == SCIF_REGISTER_FAILED )
            {
                return COI_ERROR;
            }

            // We'll need to tell them OUR offset. If we really wanted to, we could consolidate
            // this SEND with the initial SEND. However, the cost of the extra one way SEND
            // compared to everything else that happens the first time we have to register our
            // medium-sized dma memory chunk makes it "meh" when considering code complexity increase.
            result = ScifSendWrapper(&m_registered_space_offset, sizeof(m_registered_space_offset));
            if( result != COI_SUCCESS )
            {
                return result;
            }
        }

        // Copy the data to a space where DMA can be used
        memcpy(m_registered_space, message_to_send.buffer(), message_size);

        // DMA to the remote offset
        // At the byte range we are in, we use the DMA engine, not the CPU to make the transfer.
        STATIC_ASSERT( COI_MAX_SEND_RECV_MESSAGE_SIZE > THRESHOLD_SCIF_WRITETO );

        status = scif_writeto(m_endpoint, m_registered_space_offset,
                    CACHELINE_CEIL(message_size), m_remote_offset, 0);

        // We just called scif_writeto or scif_vwriteto. Either way, we do error handling and then fence
        if(status == -1)
        {
            if( IsErrnoCOI_PROCESS_DIED(errno) )
            {
                result = COI_PROCESS_DIED;
            }
            else
            {
                result = COI_ERROR;
            }
            goto send_dma_status1;
        }

        result = COIDMAFence::ScifFenceLocalWrapper(
                    m_endpoint, CACHELINE_CEIL(message_size),
                    m_send_signal_page, m_send_signal_offset);
        if(result != COI_SUCCESS)
        {
            status = -1;
            goto send_dma_status1;
        }

send_dma_status1:

        header->scif_write_status = status;

        // Tell the receiver if we were able to complete the DMA.
        if( status == 0 )
        {
            result = ScifSendWrapper(header, sizeof(Header_t));
        }
        else
        {
            (void) ScifSendWrapper(header, sizeof(Header_t));
        }


        if( result != COI_SUCCESS )
        {
            return result;
        }

        // Block until we get the "we are done memcpy" msg
        result = ScifRecvWrapper(&status, sizeof(status));

        return result;
    }
    else
    {
        // if the message is not small, the current layout of the data looks like
        // {
        //      uint64_t size; //aligned to PAGE_SIZE
        //      char     empty_unused_space[PAGE_SIZE-sizeof(uint64_t)];
        //      char     rest_of_data[]; //obtainable via .buffer(), of size size()
        // }
        //
        // This is the same as the previous size, but the difference here will be to use
        // scif_vwriteto instead of scif_writeto + the temporary already-registered DMA buffer + memcpy approach

        // Allocate the signal page if necessary and register it.
        // Either both succeed or both must be initialized to their "fail" states
        if( message_size <= COIDMAFence::MAXSPINSIZE && m_send_signal_offset == SCIF_REGISTER_FAILED )
        {
            result = AllocateSignalPage();
            if( result != COI_SUCCESS )
            {
                return result;
            }
        }

        // Let the other side know about the message
        result = ScifSendWrapper(message_to_send.rawdata(), COI_MIN_SEND_RECV_MESSAGE_SIZE);
        if( result != COI_SUCCESS )
        {
            return COI_ERROR;
        }

        off_t remote_offset = SCIF_REGISTER_FAILED;

        // Receive the offset that the receiver registered
        result = ScifRecvWrapper( &remote_offset, sizeof(remote_offset));

        DPRINTF("ScifRecvWrapper %lu bytes returned %d. The offset received"
                " 0x%lx.  errno = %d\n",
                sizeof(remote_offset), result,
                remote_offset, errno);
        if (result != COI_SUCCESS)
        {
            return result;
        }
        if (remote_offset == SCIF_REGISTER_FAILED)
        {
            return COI_ERROR;
        }

        DPRINTF("About to vwriteto on destination %d and offset 0x%lx with "
                "length %lu\n", m_endpoint, remote_offset,
                message_size);

        // At the byte range we are in, we use the DMA engine, not the CPU to make the transfer.
        STATIC_ASSERT( COI_MAX_SEND_RECV_MESSAGE_SIZE > THRESHOLD_SCIF_VWRITETO );

        status = scif_vwriteto( m_endpoint, message_to_send.buffer(),
                                CACHELINE_CEIL(message_size), remote_offset,
                                0);

        // error handling and then fence
        if(status == -1)
        {
            if( IsErrnoCOI_PROCESS_DIED(errno) )
            {
                result = COI_PROCESS_DIED;
            }
            else
            {
                result = COI_ERROR;
            }
            goto send_dma_status2;
        }

        result = COIDMAFence::ScifFenceLocalWrapper(
                    m_endpoint, CACHELINE_CEIL(message_size),
                    m_send_signal_page, m_send_signal_offset);
        if(result != COI_SUCCESS)
        {
            status = -1;
            goto send_dma_status2;
        }

send_dma_status2:

        // Send the message to the sink telling it you are done copying data
        // or that there was an error.
        if( status == 0 )
        {
            result = ScifSendWrapper(&status, sizeof(status));
        }
        else
        {
            (void) ScifSendWrapper(&status, sizeof(status));
        }

        if (result != COI_SUCCESS)
        {
            return result;
        }

        return result;
    }
}

// Receive a message. If it returns successfully, this function will ALWAYS
// Allocate() the buffer needed for out_message_to_recv.
// The Message_t class will handle freeing any space allocated here.
COIRESULT
_SCIFComm::ReceiveUnsafe(Message_t& out_message_to_recv)
{
    if (!m_initialized)
    {
        return COI_NOT_INITIALIZED;
    }

    if (m_endpoint == SCIF_OPEN_FAILED)
    {
        return COI_OUT_OF_RANGE;
    }

    COIRESULT result = COI_ERROR;

    // Always scif_recv the first few bytes.
    // These will tell us how many more bytes need to be received.
    Header_t header;
    result = ScifRecvWrapper(&header,
                             COI_MIN_SEND_RECV_MESSAGE_SIZE);

    // If an error occurred
    if (result != COI_SUCCESS)
    {
        DPRINTF("Result = %d\n", (result));
        return result;
    }

    const uint64_t message_size = header.size;
    const uint64_t raw_size = message_size + sizeof(uint64_t);

    DPRINTF("RecvMessage %d Allocating %lu bytes\n",
                m_endpoint, message_size);

    out_message_to_recv.Allocate(message_size);

    if (raw_size <= COI_MIN_SEND_RECV_MESSAGE_SIZE)
    {
        // if the message was small then we are done. just memcpy the data
        memcpy(out_message_to_recv.buffer(),
                header.data,
                COI_HEADER_DATA_SIZE);
        result = COI_SUCCESS;
    }
    else if (raw_size <= COI_MAX_SEND_RECV_MESSAGE_SIZE)
    {
        // We've already received COI_MIN_SEND_RECV_MESSAGE_SIZE.
        // Let's go ahead and copy the first portion
        memcpy(out_message_to_recv.buffer(),
                        header.data,
                        COI_HEADER_DATA_SIZE);

        // They still need to scif_send more data, let's scif_receive it and
        // store it in the appropriate offset.
        result = ScifRecvWrapper(out_message_to_recv.buffer() + COI_HEADER_DATA_SIZE,
                                 message_size - COI_HEADER_DATA_SIZE);

    }
    else if( message_size <= COI_MAX_REGISTERED_MESSAGE_SIZE )
    {
        // We are in weird place where it would be slow for the sender to scif_vwriteto
        // data because of the registration cost each time.
        // Instead we'll register some memory we have set aside and use it
        // as a temporary place to DMA stuff to/from.
        // After the DMA is complete, then memcpy that into the output message.
        // There is some additional "handshake"/verification stuff that needs
        // to happen, but all that is STILL faster then using scif_send/recv
        // or to scif_v* transfer it. Of course, this will vary depending on the actual
        // host/device specs, and you can use heuristics to change the appropriate size
        // to hit the "sweet spot" for which there are performance gains here.
        // The recommendation is to optimize the case for run functions with all of the
        // misc data used up plus some additional space for overhead involved in requesting
        // a run function.
        int remote_status = header.scif_write_status;

        // If this is the first time for a transfer in this size range
        // there's some handshaking to setup the temporary DMA buffers
        // and exchange scif offsets with the sender.
        if( m_registered_space_offset == SCIF_REGISTER_FAILED )
        {
            // Allocate and scif_register some memory
            result = AllocateRegisteredMemory();

            // Send our offset first
            if( result != COI_SUCCESS )
            {
                (void) ScifSendWrapper(&m_registered_space_offset, sizeof(m_registered_space_offset));
            }
            else
            {
                result = ScifSendWrapper(&m_registered_space_offset, sizeof(m_registered_space_offset));
            }

            if(result != COI_SUCCESS)
            {
                return result;
            }

            // If we just allocated stuff that they just did too
            // and we'll also need their offset.
            result = ScifRecvWrapper(&m_remote_offset, sizeof(m_remote_offset));
            if( result != COI_SUCCESS )
            {
                return result;
            }
            if( m_remote_offset == SCIF_REGISTER_FAILED )
            {
                return COI_ERROR;
            }

            // This is the first transfer for this size range.
            // It works slightly different than future transfers and just need to
            // wait on the "we are done writing to registered memory" message
            Header_t remote_status_header;
            result = ScifRecvWrapper(&remote_status_header, sizeof(remote_status_header));
            remote_status = remote_status_header.scif_write_status;

            if( result != COI_SUCCESS )
            {
                return result;
            }
            if( remote_status != 0 )
            {
                return COI_ERROR;
            }
        }
        // Else the sender had everything needed to DMA the data over. We just
        // need to be notified when the sender has completed the DMA so we can do the memcpy.

        // We know that the sender successfully wrote the data. We need to copy it to
        // the outgoing message now
        memcpy(out_message_to_recv.buffer(), m_registered_space, message_size);

        // And finally, we notify the sender that we are done with the copy
        // to protect them from overwriting the temporary buffer.
        result = ScifSendWrapper(&remote_status, sizeof(remote_status));
        return result;
    }
    else
    {
        // The sender will use scif_vwriteto to send data.
        // The memory is a "one-time" transfer, so we'll need to register
        // an appropriate memory location to receive that data, send the
        // offset back to them, and then wait for confirmation that they
        // completed the write.

        // register the memory window
        off_t window_offset = SCIF_REGISTER_FAILED;
        int remote_status = -1;
        int status = -1;
        DPRINTF("About to register on endpoint %d\n", (int)m_endpoint);
        errno = 0;
        window_offset = scif_register(m_endpoint, out_message_to_recv.buffer(),
            PAGE_CEIL(message_size), 0, SCIF_PROT_WRITE, 0);

        DPRINTF("scif_register of address %lu and size %lu returned %lu aka "
            "0x%lx. errno = %d\n", (uint64_t)out_message_to_recv.buffer(),
            PAGE_CEIL(message_size), window_offset, window_offset, errno);

        // Send the window offset from the register call, even if the register
        // call failed
        result = ScifSendWrapper(&window_offset, sizeof(window_offset));

        // If there was a problem registering the memory or sending the window
        // we shouldn't be receiving messages so we can exit early
        if (window_offset == SCIF_REGISTER_FAILED)
        {
            DPRINTF("ERROR in the scif_register\n");
            return COI_ERROR;
        }
        if( result != COI_SUCCESS )
        {
            goto unregister;
        }

        // We can now wait on the "we are done writing to registered memory"
        // message

        result = ScifRecvWrapper(&remote_status,
                                 sizeof(remote_status));
        if (result != COI_SUCCESS)
        {
            DPRINTF("Error receiving the message from the sender.\n");
            goto unregister;
        }
        if (remote_status != 0)
        {
            DPRINTF("Sender unable to vwriteto or fence."
                " remote status = %d\n", remote_status);
            result = COI_ERROR;
            goto unregister;
        }

unregister:
        // Sender has finished copying the data or there was an error.
        // Let's unregister the space.
        status = scif_unregister(m_endpoint, window_offset,
            PAGE_CEIL(message_size));
        if (status != 0 && result == COI_SUCCESS)
        {
            result = COI_ERROR;
        }
    }
    DPRINTF("RecvMessage %d END with COIRESULT %d \n", m_endpoint, result);
    return result;

}

COIRESULT
_SCIFComm::SendMessageAndReceiveResponseUnsafe(Message_t& msg_to_send, Message_t& msg_to_recv)
{
    COIRESULT send_result = SendUnsafe( msg_to_send );
    if( send_result == COI_SUCCESS )
    {
        return ReceiveUnsafe( msg_to_recv );
    }
    return send_result;
}

//=============================================================================
// _SCIFComm Class -- End
//=============================================================================


