/*
 * Copyright 2010-2013 Intel Corporation.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, version 2.1.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Disclaimer: The codes contained in these modules may be specific
 * to the Intel Software Development Platform codenamed Knights Ferry,
 * and the Intel product codenamed Knights Corner, and are not backward
 * compatible with other Intel products. Additionally, Intel will NOT
 * support the codes or instruction set in future products.
 *
 * Intel offers no warranty of any kind regarding the code. This code is
 * licensed on an "AS IS" basis and Intel is not obligated to provide
 * any support, assistance, installation, training, or other services
 * of any kind. Intel is also not obligated to provide any updates,
 * enhancements or extensions. Intel specifically disclaims any warranty
 * of merchantability, non-infringement, fitness for any particular
 * purpose, and any other warranty.
 *
 * Further, Intel disclaims all liability of any kind, including but
 * not limited to liability for infringement of any proprietary rights,
 * relating to the use of the code, even if Intel is notified of the
 * possibility of such liability. Except as expressly stated in an Intel
 * license agreement provided with this code and agreed upon with Intel,
 * no license, express or implied, by estoppel or otherwise, to any
 * intellectual property rights is granted herein.
 */

/// @file sysinfo_common.cpp

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <map>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdint.h>

#include <sys/types.h>
#include <scif.h>



#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "common/COISysInfo_common.h"
#include "common/COIEngine_common.h"
#include "source/COIEngine_source.h"

#include "internal/_SysInfo.h"
#include "internal/_Debug.h"

_COISysInfo::proc_t*    _COISysInfo::s_procs = NULL;
uint32_t                _COISysInfo::s_procs_len = 0;

// Mapping from API to logical processor. This array is as long as the highest
// 'apicid' we find and may contain NULL elements if an apicid doesn't have a
// mapping (I don't think this is possible, but we make no assumptions).
_COISysInfo::proc_t**   _COISysInfo::s_apic_map = NULL;
uint32_t                _COISysInfo::s_apic_map_len = 0;
uint32_t                _COISysInfo::s_total_sockets = 0;
uint32_t                _COISysInfo::s_total_cores = 0;

// Reads lines from /proc/cpuinfo until it finds one that starts with a
// given key. We are expecting lines such as:
//   processor       : 1
//   vendor_id       : GenuineIntel
//
// E.g. given "cpu family", we skip to through lines until we find.
//   cpu family      : 11
//                     ^ STOP HERE
bool _COISysInfo::SkipToKeyValue(FILE *f, const char *key)
{
    // Skip over lines until we find the key
    int n = 0;
    char chr;
    while (key[n] != 0) {
        chr = fgetc(f);
        if (key[n] == chr) {
            n++;
        } else {
            // prefix mismatch, skip to EOL to try next line
            while (chr != '\n') {
                if (chr == EOF) {
                    return false;
                }
                chr = fgetc(f);
            }
            n = 0;
        }
    }

    // Skip over the whitespace until we hit the ':'
    while ((chr = fgetc(f)) != ':') {
        if (chr == EOF || chr == '\n') {
            return false;
        }
    }

    // skip the space(s?) after the ':'
    while (isspace((chr = fgetc(f)))) {
        if (chr == EOF || chr == '\n') {
            return false;
        }
    }
    // stdio.h guarantees at least one character pushback for ungetc
    ungetc(chr,f);

    return true;
}


// Load all sorts of processor info once and cache the data in the static
// s_ variables.
bool _COISysInfo::LoadProcessorInfo()
{
	static pthread_mutex_t   s_sysinfoLock = PTHREAD_MUTEX_INITIALIZER;
    static bool              s_init = false;

    PT_ASSERT(pthread_mutex_lock(&s_sysinfoLock));
    if (s_init) {
        PT_ASSERT(pthread_mutex_unlock(&s_sysinfoLock));
        return true;
    }

    FILE *f = fopen("/proc/cpuinfo","r");
    if (f == NULL) {
        fprintf(stderr, "LoadProcessorInfo(): failed to open /proc/cpuinfo\n");
        PT_ASSERT(pthread_mutex_unlock(&s_sysinfoLock));
        return false;
    }

    uint32_t procs_len = 0, procs_cap = 16;
    proc_t *procs = (proc_t *)malloc(procs_cap * sizeof(*procs));
    if (procs == NULL) {
        PT_ASSERT(pthread_mutex_unlock(&s_sysinfoLock));
        fclose(f);
        return false;
    }

    uint32_t max_apic = 0;
    uint32_t nextCoreIX = 0, nextSockIX = 0;

    std::map<uint32_t,uint32_t> core_ids, socket_ids;
    std::map<uint32_t,uint32_t>::iterator itr;

    // Run through the keys of /proc/cpuinfo for each logical processor (hw
    // thread). From this we look at various keys to determine the actual
    // topology.
    uint32_t procID;
    while (FindUInt32(f, "processor", &procID)) {

#define LP_PARSE_KEY(t,k,p) \
    { \
        if (!Find ## t (f,k,p)) { \
            fprintf(stderr, "%s:%d: failed to parse key %s in /proc/cpuinfo " \
                "for processor %d\n", __FILE__, __LINE__, k, procs_len); \
            fclose(f); \
            free(procs); \
            return false; \
        } \
    }

        double cpuMHZ;
        uint32_t sockID, coreID, apicID;

        LP_PARSE_KEY(Double, "cpu MHz", &cpuMHZ);
        LP_PARSE_KEY(UInt32, "physical id", &sockID);
        LP_PARSE_KEY(UInt32, "core id", &coreID);
        LP_PARSE_KEY(UInt32, "apicid", &apicID);

#undef LP_PARSE_KEY

        max_apic = apicID > max_apic  ? apicID : max_apic;

        // Translate 'core id' to core index. Core ids are not logically
        // ordered [0,n-1] and may skip ranges of integers. We impose such
        // a natural number ordering.
        uint32_t coreIX;
        itr = core_ids.find(coreID);
        if (itr == core_ids.end()) {
            coreIX = core_ids[coreID] = nextCoreIX++;
        } else {
            coreIX = itr->second;
        }

        // Translate 'physical id' to socket index. Similar to core ids, we
        // may have to map them to a natural order.
        uint32_t sockIX;
        itr = socket_ids.find(sockID);
        if (itr == socket_ids.end()) {
            sockIX = socket_ids[sockID] = nextSockIX++;
        } else {
            sockIX = itr->second;
        }

        if (procs_len++ >= procs_cap) {
            procs_cap += 16;
            proc_t *p = (proc_t *)realloc(procs, procs_cap * sizeof(*procs));
            if (p == NULL) {
                goto fail;
            }
            procs = p;
        }
        proc_t *p = &procs[procs_len - 1];
        p->m_procID = procID;
        p->m_sockID = sockIX;
        p->m_coreID = coreIX;
        p->m_procID = procID;
        p->m_apicID = apicID;
        // printf("%p: p%d: c%d, s%d, a%d\n",
        //     p, procID, coreID, sockID, apicID);

        // If the kernel allows the cpu to throttle, 'cpu MHz' will be
        // inaccurate since it reflects the current frequency, not the max.
        // Hence, we first try and get the correct max value the sysfs file
        // system.
        char fname[128];
        snprintf(fname, sizeof(fname),
            "/sys/devices/system/cpu/cpu%d/cpufreq/cpuinfo_max_freq", procID);
        // This is indexed by logical cpu number [2].
        //
        // N.B. if you go poking around in this directory structure you'll
        // find a .../cpu#/topolgy/core_id file, this core id may differ
        // from the 'core id' field in /proc/cpuinfo for the same logical CPU
        // since this is the hardware id for the core, not the OS's id for
        // that core. Hence, it may differ (seee [2]).
        FILE *freq_file = fopen(fname, "r");
        if (freq_file != NULL) {
            if (fscanf(freq_file,"%ld",&p->m_maxFreqHz) != 1) {
                fprintf(stderr, "failed to read freq from %s", fname);
                fclose(freq_file);
                goto fail;
            }
            // The value is given in kHz, convert to hz (see [3]).
            p->m_maxFreqHz *= 1000;
            fclose(freq_file);
        } else {
            // If the path fname doesn't exist, then the cpufreq module is
            // not loaded. This implies that the CPU is no throttled and
            // should be running at full speed. Thus the 'cpu MHz' field is
            // the true maximum speed and we can safely use it.
            p->m_maxFreqHz = (uint64_t) (1000.0 * 1000.0 * cpuMHZ);
        }
    }

    s_procs = procs;
    s_procs_len = procs_len;

    // count the unique socket id's
    s_total_sockets = nextSockIX;
    // The total number of cores.
    s_total_cores = nextCoreIX;

    // Create the apic map.
    s_apic_map_len = max_apic + 1;
    s_apic_map = (proc_t **)calloc(s_apic_map_len, sizeof(*s_apic_map));
    if (!s_apic_map)
    {
        goto fail;
    }
    for (uint32_t pi = 0; pi < procs_len; pi++) {
        s_apic_map[s_procs[pi].m_apicID] = &s_procs[pi];
    }

    s_init = true;

    if (f != NULL) {
        fclose(f);
    }
    PT_ASSERT(pthread_mutex_unlock(&s_sysinfoLock));
    // DumpProcessorTopology();
    return true;
fail:
    if (f != NULL) {
        fclose(f);
    }
    free(procs);
    PT_ASSERT(pthread_mutex_unlock(&s_sysinfoLock));
    return false;
}

uint64_t _COISysInfo::GetHardwareThreadMaxFrequency(uint32_t ti)
{
    if (!LoadProcessorInfo() || ti >= s_procs_len)
    {
        return 0;
    }
    return s_procs[ti].m_maxFreqHz;
}

// Given a scif node#, return the string representation of a particular node.
// We assume users followed the guidance and kept aliases to their devices:
// host, mic0, mic1, etc...
// Otherwise there is no reliable way to determine the IP addr.
const char* _COISysInfo::GetDeviceIpAddress(uint16_t node)
{
    // Thread safe: no locking necessary. If two thread race to create it,
    // it won't matter, same data gets written
    static bool     initialized = false;
    static char     ip_addresses[MAX_DEVICES + 1][MAX_ADDR_LEN];
    static uint16_t num_nodes;

    if (!initialized)
    {
        uint16_t self = 0;
        uint16_t nodes[MAX_DEVICES + 1];
        bool err = false;

        if ((num_nodes = scif_get_nodeIDs(nodes, MAX_DEVICES + 1, &self)) ==
            (uint16_t)-1)
        { // Should never fail
            fprintf(stderr, "DeviceIpAddress: scif_get_nodeIDs\n");
            return NULL;
        }

        // Special treatment for device 0
        if (self == 0)
        {
            // Host: use the loopback address
            strncpy(ip_addresses[0],"127.0.0.1",sizeof(ip_addresses[0]) - 1);
            ip_addresses[0][sizeof(ip_addresses[0]) - 1] = '\0';
        }
        else
        {
            // Device: try and lookup the host name "host"
            if (!lookup_address("host", ip_addresses[0]))
            {
                // failing that, use 172.31.1.254 
                strncpy(ip_addresses[0],"172.31.1.254",sizeof(ip_addresses[0]) - 1);
                ip_addresses[0][sizeof(ip_addresses[0]) - 1] = '\0';
                err = true;
            }
        }

        if (num_nodes >= MAX_DEVICES + 1)
        {
            fprintf(stderr, "DeviceIpAddress: too many nodes\n");
            return NULL;
        }

        // Look for mic0, mic1, etc.
        for (int i = 1; i < num_nodes; i++)
        {
            char dev_name[16] = {0};
            
            snprintf(dev_name, 16, "mic%d", i-1);
            if (!lookup_address(dev_name, ip_addresses[i]))
            {
                err = true;
                continue;
            }
        }

        initialized = !err;
    }

    // Strictly GT because there is a node 0.
    if (node > num_nodes)
    {
        errno = EINVAL;
        return NULL;
    }

    return ip_addresses[node];
}

